""" webfilter.__init__.py

    Copyright (C) 2000,2001  Bastian Kleineidam

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Welcome to the wonderful world of software OSI layer 5 filtering.
If you want to write your own filter look at the example filters.
Because several threads are using the filter code, be sure to be
thread-safe! You can add extern states for the filter by making
a separate class and passing it with the "attrs" parameter. Look
at the Rewriter filter to see how its done.
"""
import sys,WebCleanerConf
from get_dom import *
_FILTERS = []
_DEBUG_LEVEL = 0
_FILTER_LIST = [[],[],[],[],[],[],[],[],[],[]]

try:
    import fintl,os
    gettext = fintl.gettext
    domain = 'webcleaner'
    localedir = os.path.join(WebCleanerConf.install_data, 'locale')
    fintl.bindtextdomain(domain, localedir)
    fintl.textdomain(domain)
except ImportError, msg:
    print msg
    def gettext(msg):
        return msg
# set _ as an alias for gettext
_ = gettext


def startfunc():
    if os.name=='posix':
        import signal
        signal.signal(signal.SIGHUP, configure)
    configure(None, None)
    import proxy,socket
    # wrap it up in a try statement because under Linux the interrupted
    # I/O function socket.accept() raises an exception:
    # socket.error: (4, 'Interrupted system call')
    while 1:
        try:
            proxy.ProxyServer(proxy._PORT).serve_forever()
        except socket.error:
            type, value = sys.exc_info()[:2]
            debug("%s (%s)" % (str(type), str(value)))
            pass


# configuration
def configure(signum, frame):
    """The signum and frame args are needed for signal.signal()
       You can fill them with (None, None) values"""
    configfile = os.path.join(WebCleanerConf.config_dir, "webcleaner.conf")
    dom = get_dom(configfile)
    import proxy
    proxy.configure(dom)
    filters = dom.getElementsByTagName('filter')
    for f in filters:
        _FILTERS.append(get_node_attr(f, 'name'))
    from glob import glob
    files = glob(WebCleanerConf.config_dir+"/*.zap")
    doms = []
    # read filter files
    for f in files:
        debug('reading filter file %s' % f)
        dom = get_dom(f)
        attrs = get_dom_attrs(dom)
        title = get_attr(attrs, 'title')
        if int(get_attr(attrs, 'disable')):
            debug("disable zapper '%s'" % title)
            continue
        lang = get_attr(attrs, 'lang')
        if lang and lang!='Python':
            debug("disable zapper '%s'" % title)
            continue
        debug("enable zapper '%s'" % title)
        doms.append(dom)
    # fill the filter table
    for f in _FILTERS:
        exec "import %s" % f
        _module = getattr(sys.modules['webfilter'], f)
        for order in getattr(_module, 'orders'):
            instance = getattr(_module, f)()
            for dom in doms:
                instance.configure(dom)
            _FILTER_LIST[order].append(instance)


# filter order
FILTER_REQUEST         = 0 # Filter complete request (blocking)
FILTER_REQUEST_HEADER  = 1 # Outgoing header manglers
FILTER_REQUEST_DECODE  = 2 # May decode outgoing content.
FILTER_REQUEST_MODIFY  = 3 # May modify outgoing content.
FILTER_REQUEST_ENCODE  = 4 # May encode outgoing content.
FILTER_RESPONSE        = 5 # Filter complete response
FILTER_RESPONSE_HEADER = 6 # Incoming header manglers
FILTER_RESPONSE_DECODE = 7 # May decode incoming content
FILTER_RESPONSE_MODIFY = 8 # May modify incoming content
FILTER_RESPONSE_ENCODE = 9 # May encode incoming content


def printFilterOrder(i):
    if i==0: return "Request"
    elif i==1: return "Request Header"
    elif i==2: return "Request Decode"
    elif i==3: return "Request Modify"
    elif i==4: return "Request Encode"
    elif i==5: return "Response"
    elif i==6: return "Response Header"
    elif i==7: return "Response Decode"
    elif i==8: return "Response Modify"
    elif i==9: return "Response Encode"
    return "Invalid"


def initStateObjects(mime="text/html", headers={}):
    attrs = {'mime': mime}
    for i in range(10):
        for f in _FILTER_LIST[i]:
            if hasattr(f, 'mimelist'):
                if mime in f.mimelist:
                    attrs.update(f.getAttrs(headers))
            else:
                attrs.update(f.getAttrs(headers))
    return attrs


def applyfilter(i, arg, fun='filter', attrs={}):
    """Apply all filters which are registered in filter level i.
    For different filter levels we have different arg objects.
    Look at the filter examples.
    An exception in a filter causes the filtering to stop
    and we return the result we have achieved so far.
    """
    try:
        debug('filter stage %s' % printFilterOrder(i), 2)
        for f in _FILTER_LIST[i]:
            ffun = getattr(f, fun)
            if hasattr(f, 'mimelist'):
                if attrs['mime'] in f.mimelist:
                    arg = apply(ffun, (arg,), attrs)
            else:
                # no mimelist? then this filter applies to all files!
                arg = apply(ffun, (arg,), attrs)
    finally:
        return arg


# debug function, using the _DEBUG_LEVEL
def debug(s, level=1):
    if level <= _DEBUG_LEVEL:
        sys.stderr.write("DEBUG: %s\n" % s)

def error(s):
    sys.stderr.write("error: %s\n" % s)

# The base filter class
class Filter:
    def configure(self, dom):
        pass

    def filter(self, data, **args):
        return data

    def finish(self, data, **args):
        return data

    def getAttrs(self, headers):
        return {}
