#####################################################################
# munchy.py : a proxy server
# Neil Schemenauer <nascheme@enme.ucalgary.ca>
#
# renamed to proxy.py and adjusted for WebCleaner by
# Bastian Kleineidam <calvin@users.sourceforge.net>
#####################################################################
import sys,time,string,socket,select,urllib,urlparse,SocketServer
import webfilter,WebCleanerConf
from webfilter import debug,_,applyfilter

# global vars used in this program
_PORT=None
_BUF_SIZE=None
_LOGFILE=None
_PARENT_PROXY=None
_PARENT_PROXY_PORT=None
_TIMEOUT=None
_OBFUSCATE_IP=None
_ERROR_TEXT = _("""<html><head>
<title>WebCleaner Proxy Error %d %s</title>
</head><body bgcolor="#fff7e5"><br><center><b>Bummer!</b><br>
WebCleaner Proxy Error %d %s<br>
%s<br></center></body></html>""")
_ERROR_LEN = len(_ERROR_TEXT)

def log(msg):
    """If _LOGFILE is defined write the msg into it. The message msg
       must be in common log file format."""
    if _LOGFILE:
        _LOGFILE.write(msg+"\n")


def split_netloc(netloc, defaultport):
    """Split the port number from the host. The default port should be
       either 80 (HTTP) or 443 (HTTPS)"""
    if ':' in netloc:
        # find port number
        host, port = string.split(netloc, ':')
        return host, int(port)
    return netloc, defaultport


def obfuscate(host):
    """change a hostname into an integer"""
    host = socket.gethostbyname(host)
    host = map(int, string.split(host, '.'))
    return str((host[0]<<24)+(host[1]<<16)+(host[2]<<8)+host[3])


class ProxyHandler(SocketServer.StreamRequestHandler):
    _orders = [webfilter.FILTER_RESPONSE_DECODE,
               webfilter.FILTER_RESPONSE_MODIFY,
               webfilter.FILTER_RESPONSE_ENCODE]

    def handle(self):
        """handle one request from the browser"""
        try:
            srfile, swfile, ssl = self.handle_request()
            self.handle_response(srfile, swfile, ssl)
        except (socket.error, IOError):
            type, value = sys.exc_info()[:2]
            debug("%s (%s)" % (str(type), str(value)))
            pass # browser closed connection?
        debug('finished request')


    def handle_request(self):
        """read request to find host and port and make new request"""
        request = self.rfile.readline()
        log('%s - %s - %s' % \
	    (self.client_address[0], time.ctime(time.time()), request))
        request = applyfilter(webfilter.FILTER_REQUEST, request)
        debug("read request %s" % `request`)
        # split the request string
        try:
            method, url, protocol = string.split(request)
        except:
            self.error(400, _("Can't parse request"))
        if not url:
            self.error(400, _("Empty URL"))
        if method not in ['GET', 'HEAD', 'POST', 'CONNECT']:
            self.error(501, _("Unknown request method (%s)") % method)
        if method=='CONNECT':
            # urlparse chokes if the scheme is missing and a port is defined
            url = "https://"+url
            ssl = 1
        else:
            ssl = 0
        # split url into site and path
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
        if string.lower(scheme) not in ['http', 'https']:
            self.error(501, _("Unknown request scheme (%s)") % scheme)
        host,port = split_netloc(netloc, scheme and 80 or 443)
        if _OBFUSCATE_IP:
            debug("cleartext host ip: %s" % host, 2)
            host = obfuscate(host)
            debug("obfuscated host ip: %s" % host, 2)
        if _PARENT_PROXY:
            host = _PARENT_PROXY
            port = _PARENT_PROXY_PORT
            path = urlparse.urlunparse((scheme,"%s:%d" % (host, port), path,
	                                params, query, fragment))
        else:
            path = urlparse.urlunparse(('','', path, params, query, fragment))
        # read headers from client
        headers = self.read_headers(self.rfile)
        applyfilter(webfilter.FILTER_REQUEST_HEADER, headers)
        if method == 'POST' and not headers.has_key('content-length'):
            self.error(400, _("Missing Content-Length for POST method"))
        length = int(headers.get('content-length', 0))
        # read content if any
        content = ''
        if length:
            content = self.rfile.read(length)
            debug('content = %s' % `content`, 2)
            content=applyfilter(webfilter.FILTER_REQUEST_DECODE, content)
            content=applyfilter(webfilter.FILTER_REQUEST_MODIFY, content)
            content=applyfilter(webfilter.FILTER_REQUEST_ENCODE, content)
        # build new request for server
        if not ssl:
            for key in ('accept-encoding','proxy-connection'):
                if headers.has_key(key):
                    del headers[key]
            request = '%s %s HTTP/1.0\r\n%s\r\n%s' % \
	          (method, path, self.join_headers(headers), content)
        # connect to server
        try:
            addr = socket.gethostbyname(host)
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            debug("connect to %s:%d" % (addr,port))
            server.connect((addr, port))
        except socket.error, err:
            self.error(400, _('Error connecting to "%s" (%s)') % \
                      (host, err))
        srfile = server.makefile('rb')
	swfile = server.makefile('wb')
        # send new request to server
        if not ssl:
            try:
                debug('sending %s to server' % `request`)
                swfile.write(request)
                swfile.flush()
            except socket.error, err:
                self.error(500,_('Error sending data to "%s" (%s)') % \
                          (host, err))
        else:
            self.request.setblocking(0)
            pass
        return (srfile, swfile, ssl)


    def proxy_default(self, method, scheme, host, port, path, url, protocol):
        # check blocking
        if string.lower(scheme) == 'file':
            debug("redirect to "+url)
            self.wfile.write('HTTP/1.0 200 OK\r\n')
            self.wfile.write('Content-type: %s\r\n' % protocol)
            self.wfile.write("\r\n")
            self.wfile.write(urllib.urlopen(url).read())
            self.wfile.flush()
            raise SystemExit


    def read_headers(self, infile):
        headers = {}
        name = ''
        while 1:
            line = infile.readline()
            debug("header line %s" % `line`, 2)
            if line == '\r\n' or line == '\n' or not line:
                break
            if line[0] in ' \t':
                # continued header
                headers[name] = headers[name] + '\r\n ' + string.strip(line)
            else:
                i = string.find(line, ':')
                assert(i != -1)
                name = string.lower(line[:i])
                if headers.has_key(name):
                    # merge value
                    headers[name] = headers[name] + ', ' + string.strip(line)
                else:
                    headers[name] = string.strip(line[i+1:])
        return headers


    def join_headers(self, headers):
        data = []
        for name, value in headers.items():
            data.append('%s: %s\r\n' % (name, value))
        return string.join(data, '')


    def handle_response(self, srfile, swfile, ssl):
        if ssl:
            response = 'HTTP/1.0 200 Connection established'
        else:
            debug('reading server response')
            response = srfile.readline()
        debug('response = %s' % `response`, 2)
        response = applyfilter(webfilter.FILTER_RESPONSE, response)
        debug('response = %s' % `response`, 2)
        fields = string.split(response)
        if len(fields)<2:
            self.error(500, _('Error reading response'))
        version = fields[0]
        status = fields[1]
        comment = string.join(fields[2:])
        response = 'HTTP/1.0 %s %s\r\n' % (status, comment)
        #debug('response = %s' % `response`, 2)
        self.wfile.write(response)
        if ssl:
            self.wfile.write('Proxy-agent: WebCleaner/%s\r\n' % WebCleanerConf.version)
        else:
            debug('reading response headers')
            headers = self.read_headers(srfile)
            headers = applyfilter(webfilter.FILTER_RESPONSE_HEADER, headers)
            mime = headers.get('content-type', 'application/octet-stream')
            # strip extension
            i = string.find(mime, ';')
            if i != -1:
                mime = mime[:i]
            # init filter attributes
            attrs = webfilter.initStateObjects(mime, headers)
            debug('writing headers %s to client' % `headers`, 2)
            self.wfile.write(self.join_headers(headers))
        self.wfile.write('\r\n')
        if ssl:
            # forward all subsequent SSL packets in both directions
            debug("forward ssl", 2)
            while 1:
                r = select.select([self.rfile, srfile], [], [], _TIMEOUT)
                if not r[0]:
                    raise socket.error,"SSL connection timed out"
                if self.rfile in r[0]:
                    debug("from client")
                    readfrom = self.rfile
                    writeto = swfile
                elif srfile in r[0]:
                    debug("from server")
                    readfrom = srfile
                    writeto = self.wfile
                debug("reading")
                r = readfrom.read(_BUF_SIZE)
                debug(`r`)
                if not r:
                    break
                debug("writing")
                writeto.write(r)
                writeto.flush()
            debug("SSL end")
        else:
            data = srfile.read(_BUF_SIZE)
            while data:
                for i in self._orders:
                    data = applyfilter(i, data, attrs=attrs)
                self.wfile.write(data)
                data = srfile.read(_BUF_SIZE)
            for i in self._orders:
                data = applyfilter(i, data, fun='finish', attrs=attrs)
            self.wfile.write(data)
            self.wfile.flush()


    def error(self, code, body):
        """send back an error message"""
        debug('error %s (%s)' % (code, body))
        import BaseHTTPServer
        response = BaseHTTPServer.BaseHTTPRequestHandler.responses[code][0]
        self.wfile.write("HTTP/1.0 %s %s\r\n" % (code, response))
        self.wfile.write("Server: WebCleaner\r\n")
        self.wfile.write("Content-type: text/html\r\n")
        self.wfile.write("Content-length: %d\r\n" % _ERROR_LEN)
        self.wfile.write("\r\n")
        self.wfile.write(_ERROR_TEXT % (code, response, code, response, body))
        self.wfile.flush()
        # prevent sending any more data
        raise SystemExit


#ServerBase = SocketServer.TCPServer
try:
    import thread
    TCPServer = SocketServer.ThreadingTCPServer
except ImportError:
    TCPServer = SocketServer.ForkingTCPServer

class ProxyServer(TCPServer):
    def __init__(self, port):
        debug('Starting proxy on port %d' % port)
        TCPServer.__init__(self, ('', port), ProxyHandler)

    def server_bind(self):
        """when interrupting the server, reuse the socket"""
        TCPServer.server_bind(self)
        import socket
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)


def configure(dom):
    """read config params from the DOM object"""
    from get_dom import *
    global _PORT,_BUF_SIZE,_PARENT_PROXY_PORT,_PARENT_PROXY,_LOGFILE,_TIMEOUT
    global _OBFUSCATE_IP
    attrs = get_dom_attrs(dom)
    version = WebCleanerConf.version
    conf_version = get_attr(attrs, 'version')
    if conf_version > version:
        raise SystemExit, _("Wrong config file version %s, I only understand <= %s") % (conf_version, version)
    _PORT = int(get_attr(attrs, 'port'))
    _BUF_SIZE = int(get_attr(attrs, 'buffersize'))
    _PARENT_PROXY_PORT = int(get_attr(attrs, 'parentproxyport'))
    webfilter._DEBUG_LEVEL = int(get_attr(attrs, 'debuglevel'))
    _PARENT_PROXY = get_attr(attrs, 'parentproxy')
    _LOGFILE = get_attr(attrs, 'logfile')
    _TIMEOUT = int(get_attr(attrs, 'timeout'))
    _OBFUSCATE_IP = int(get_attr(attrs, 'obfuscateip'))
    if _LOGFILE == 'stdout':
        _LOGFILE = sys.stdout
    elif _LOGFILE:
        _LOGFILE = open(_LOGFILE, 'a')

